"""Functions for applying FIR filters to dask arrays.

To create a fir filter operating on dask arrays, use the `make_filter_fir_dask` function.
"""

from __future__ import annotations

from typing import Callable, NewType, TypeAlias

import dask
import dask.array as da
import numpy as np

from lisainstrument.fir_filters_numpy import (
    DefFilterFIR,
    EdgeHandling,
    FIRCoreOp,
    NumpyArray1D,
    SemiLocalMapType,
    make_numpy_array_1d,
    semilocal_map_numpy,
)

DaskArray1D = NewType("DaskArray1D", da.Array)


def make_dask_array_1d(x: da.Array) -> DaskArray1D:
    """Check that dask array is 1D or raise exception

    This improves static type checking by promoting array dimensionality to a type.
    """
    if (dims := len(x.shape)) != 1:
        msg = f"Expected dask array with one dimension, got {dims}"
        raise ValueError(msg)
    return DaskArray1D(x)


def merge_small_chunks(sizes: tuple[int, ...], threshold: int) -> tuple[int, ...]:
    """Recursively merge chunks below threshold size with neighbors

    Arguments:
        sizes: list of chunk sizes
        threshold: minimum required chunk size

    Returns:
        List of chunk sizes after recursive merge
    """

    if len(sizes) < 2:
        return sizes
    sz = list(sizes)

    for i in range(len(sz) - 1):
        if max(sz[i], sz[i + 1]) < threshold:
            m = sz[:i] + [sz[i] + sz[i + 1]] + sz[i + 2 :]
            return merge_small_chunks(tuple(m), threshold)
    for i in range(len(sz) - 1):
        if min(sz[i], sz[i + 1]) < threshold:
            m = sz[:i] + [sz[i] + sz[i + 1]] + sz[i + 2 :]
            return merge_small_chunks(tuple(m), threshold)

    return sizes


def semilocal_map_dask(
    op: SemiLocalMapType,
    bound_left: EdgeHandling,
    bound_right: EdgeHandling,
    data: DaskArray1D,
) -> DaskArray1D:
    """Apply a semi local map to dask array and employ boundary conditions.

    Arguments:
        op: the semi local mapping
        bound_left: Boundary treatment on left side
        bound_right: Boundary treatment on right side
        data: the 1D array to be mapped

    Returns:
        The mapped data. The size is the same as the input if both boundary
        conditions are ZEROPAD. A boundary condition VALID reduces the output
        size by the corresponding margin size of the semilocal map.
    """

    if op.margin_left < 0 or op.margin_right < 0:
        msg = (
            f"semilocal_map_dask: mappings with negative"
            f" margin not supported (got left={op.margin_left}, right={op.margin_right})"
        )
        raise RuntimeError(msg)

    margin = max(op.margin_left, op.margin_right)

    def whole_op(d: np.ndarray) -> np.ndarray:
        return semilocal_map_numpy(op, bound_left, bound_right, make_numpy_array_1d(d))

    if data.size < margin:
        delay_op = dask.delayed(whole_op)
        nsize = data.size
        if bound_left == EdgeHandling.VALID:
            nsize -= op.margin_left
        if bound_right == EdgeHandling.VALID:
            nsize -= op.margin_right

        res = da.from_delayed(delay_op(data), (nsize,), data.dtype)
        return make_dask_array_1d(res)

    def ext_op(d: np.ndarray) -> np.ndarray:
        r: np.ndarray = op(make_numpy_array_1d(d))
        if op.margin_left < margin:
            r = r[margin - op.margin_left :]
        if op.margin_right < margin:
            r = r[: -(margin - op.margin_right)]
        return r

    if min(data.chunks[0]) < margin:
        newsizes = merge_small_chunks(data.chunks[0], margin)
        data = data.rechunk(chunks={0: newsizes})

    ext = da.map_overlap(ext_op, data, depth={0: margin}, boundary=0, trim=False)

    if bound_left == EdgeHandling.VALID and op.margin_left > 0:
        ext = ext[op.margin_left :]
    if bound_right == EdgeHandling.VALID and op.margin_right > 0:
        ext = ext[: -op.margin_right]

    return make_dask_array_1d(ext)


FilterFirDaskType: TypeAlias = Callable[[da.Array], DaskArray1D]


def filter_fir_dask(
    fdef: DefFilterFIR,
    bound_left: EdgeHandling,
    bound_right: EdgeHandling,
    data: da.Array,
) -> DaskArray1D:
    """Apply FIR filter to 1D dask array

    Arguments:
        fdef: The definition of the FIR filter
        bound_left: Boundary treatment on left side
        bound_right: Boundary treatment on right side
        data: The 1D dask array to be filtered

    Returns:
        Filtered dask array

    """
    fmap = FIRCoreOp(fdef)
    return semilocal_map_dask(fmap, bound_left, bound_right, make_dask_array_1d(data))


def make_filter_fir_dask(
    fdef: DefFilterFIR, bound_left: EdgeHandling, bound_right: EdgeHandling
) -> FilterFirDaskType:
    """Create a function that applies a given FIR filter to dask arrays,
    employing the specified boundary treatment.

    Arguments:
        fdef: The definition of the FIR filter
        bound_left: Boundary treatment on left side
        bound_right: Boundary treatment on right side

    Returns:
        Function which accepts a single 1D dask array as input and returns
        the filtered dask array.
    """

    fmap = FIRCoreOp(fdef)

    def op(data: da.Array) -> DaskArray1D:
        return semilocal_map_dask(
            fmap, bound_left, bound_right, make_dask_array_1d(data)
        )

    return op


def numpyfy_dask(
    dafunc: Callable[[da.Array], DaskArray1D], chunks: int
) -> Callable[[np.ndarray], NumpyArray1D]:
    """Convert function operating on dask arrays to work with numpy arrays.

    The dask function will be called with a chunked dask array created from
    the numpy array, and then evaluate the result.

    Arguments:
        dafunc: function operating on a dask array
        chunks: chunk size to be used internally
    Returns:
        function operating on numpy arrays
    """

    def func(x):
        d = da.from_array(x, chunks=chunks)
        r = dafunc(make_dask_array_1d(d)).compute()
        return make_numpy_array_1d(r)

    return func
